from warnings import warn
from bestconfig.converters import *

"""Тип, которым может быть значение конфига"""
ConfigType = t.Union[str, int, float, bool, dict, list]

"""
Используется по умолчанию, при
config.get()
Может быть переопределен параметром 'cast'
"""
default_converter = UniversalConverter()


class ConfigProvider(dict):
    """
    Основной класс, с которым имеет дело пользователь
    после создания конфига.
    Предоставляет универсальный интерфейс для хранящихся данных
    Пример:
    config = Config()
    assert isinstance(config, ConfigProvider)
    print(config.logger.format)
    print(config['logger.format'])
    print(config['logger']['format'])
    print(config.logger['format'])
    print(config.get('logger'))
    """

    def __init__(self, data: dict):
        self._data: dict = data
        self._modified: bool = False
        super().__init__(data)

    def get(self, item: str, default_value=None, raise_absent=False,
            cast: t.Optional[AbstractConverter] = default_converter) -> ConfigType:
        """
        Обращается к self._data и возвращает собственный экземпляр, если значение dict
        в противном случае само значение

        attr_name = config.get('attr_name', 123)

        :param item: ключ, значение по которому нужно вернуть
        :param default_value: при отсутствии ключа, возвращается оно
        :param raise_absent: если ключа не существует, кинуть KeyError,
        в противном случает вернуть None
        :param cast: наследник класса AbstractConverter,
        если передан, возвращается cast.cast(value)
        :return: значение по ключу, None или KeyError
        """
        assert isinstance(item, str), 'Key must be str, not %s' % type(item)

        try:
            # Обработка случая config.get('key.other')
            value = self._unsafe_access_key(item)

            # Возвращаем словарь в виде класса ConfigProvider
            if isinstance(value, dict):
                return self.__class__(self._data[item])

            # Преобразуем объект в соответствии с переданным в параметрах cast
            if cast:
                return cast.cast(value)
            return value
        except KeyError:
            if raise_absent:
                raise KeyError

        return default_value

    def get_raw(self, item: str):
        """Возвращает значение по ключу, не изменяя его
        Такое, какое было считано из файла"""
        return self.get(item, cast=None)

    def assert_contains(self, item: str):
        """Бросает исключение KeyError, если ключ не найден"""
        self.get(item, raise_absent=True)

    def set(self, item: str, value: ConfigType):
        """Устанавливает значение по ключу"""
        if not item:
            warn('Использование пустой строки в качестве ключа', UserWarning)
        self._modified = True
        self._data[item] = value

    def to_dict(self) -> dict:
        """Возвращает весь конфигурационные словарь, содержащий имеющиеся данные
        Эквивалентно dict(config)"""
        return self._data

    def int(self, item: str) -> t.Optional[int]:
        """ config.int('limit') -> 45 """
        return self.get(item, cast=SimpleConverter(int))

    def float(self, item: str) -> t.Optional[float]:
        """ type(config.float('wait_time')) in [float, None] """
        return self.get(item, cast=SimpleConverter(float))

    def bool(self, item: str) -> t.Optional[float]:
        """ config.bool('DEBUG') in [True, False, None] """
        # Тут используется не универсальный преобразователь
        return self.get(item, cast=BoolConverter())

    def list(self, item: str) -> t.Optional[list]:
        """ config.list('items') -> [1, 3, 4] """
        return self.get(item, cast=PythonicConverter(empty_as=[]))

    def dict(self, item: str) -> t.Optional[dict]:
        """ config.dict('logger') -> {'mode': 'WARNING'} """
        return self.get(item, cast=PythonicConverter(empty_as=dict()))

    def str(self, item: str) -> t.Optional[dict]:
        return self.get(item, cast=SimpleConverter(str))

    def __getattr__(self, item):
        """attr_name = config.attr_name"""
        return self.get(item, raise_absent=True)

    def __getitem__(self, item):
        """attr_name = config['attr_name']"""
        return self.get(item, raise_absent=True)

    def __len__(self):
        """
        :return: количество индексируемых переменных
        """
        return len(self.__dict__)

    def _unsafe_access_key(self, item: str) -> t.Optional[ConfigType]:
        """Возвращает значение из _data, пытаясь его найти
        по строке виде key.subkey.otherkey или без точки
        Пример: config.get('logger.mode')
        config.get('ADMIN_ID')
        Кидает KeyError, при отсутствии ключа
        """

        if item in self._data:
            return self._data[item]

        # Попытка пройти по частям ключей, разделенным точками
        keys = item.split('.')
        value = self._data
        for key in keys:
            if not hasattr(value, '__getitem__'):
                raise KeyError

            if key not in value:
                raise KeyError

            value = value[key]

        return value



import os
import traceback
import typing as t
from pathlib import Path

from bestconfig.adapters import EnvAdapter, FileAdapter, DictAdapter
from bestconfig.config_provider import ConfigProvider
from bestconfig.source import Source, SourceType

supported_extensions = ['json', 'yaml', 'ini', 'cfg', 'env']
applicant_files = ['config', 'conf', 'setting', 'settings', 'configuration']


def generate_sources(files: t.List[str], extensions: t.List[str]) -> t.List[str]:
    """Генерирует всевозможные названия файлов
    на основе предоставленных претендентов на названия и расширения"""
    return list(sum(
        [
            [f'{filename}.{ext}' for ext in extensions]
            for filename in files
        ],
        []
    ))


class Config:
    """
    Основной класс, является частью публичного
    интерфейса библиотеки.

    Пример использования:

    from bestconfig import Config
    config = Config()
    print(config.limit)
    """

    def __new__(cls, *args, exclude_default=False, raise_on_absent=False, exclude: list = None) -> ConfigProvider:
        # Добавить значения по умолчанию
        sources = cls._get_sources(*args, exclude_default=exclude_default, exclude=exclude or set())
        # Передаем файл, из которого был совершен вызов Config()
        # в объекте traceback.extract_stack, последний вызов это данная функция, а перед ним, вызывающая
        caller_path = traceback.extract_stack()[-2][0]
        sources = SourcesFilter.filter(sources, caller_path=caller_path)
        aggregator = ConfigAggregator(sources)
        config_dict = aggregator.to_dict()
        return ConfigProvider(config_dict)

    """Начало и конец списка источников конфигов, те, что ближе к концу 
    кри коллизии перезаписывают более ранние"""
    _sources_begin = generate_sources(applicant_files, supported_extensions)
    _sources_end = ['.env', 'env_file', Source.env]

    @classmethod
    def _get_sources(cls, *args, exclude_default, exclude: list) -> t.List[Source]:
        """Дополняет переданные пользователем источники (файлы) стандартными"""
        exclude_set = set(exclude)
        sources = []
        if not exclude_default:
            sources += [source_name for source_name in cls._sources_begin if source_name not in exclude_set]
        sources += args
        if not exclude_default:
            sources += [source_name for source_name in cls._sources_end if source_name not in exclude_set]

        return [Source(source) for source in sources]


class SourcesFilter:
    """Делает preprocessing входных источников,
    превращает их в реальные пути к файлам, удаляет невалид и прочее"""

    @classmethod
    def filter(cls, sources: t.List[Source], caller_path: str) -> t.List[Source]:
        """Основной метод класса
        Заменяет названия файлов на реальные пути в формате Path"""
        clear_sources = []
        scanner = FilesScanner(caller_path=caller_path)
        for source in sources:
            if SourceType.FILE == source.source_type:
                found_files = scanner.find_all_files(source.data['filename'])
                for filename in found_files:
                    clear_sources.append(Source(filename))
            else:
                clear_sources.append(source)

        return clear_sources


class FilesScanner:
    """
    Задача класса находить файлы конфигурации
    в директориях проекта
    """

    def __init__(self, caller_path: str):
        self._root_path = Path(os.getcwd())
        self._caller_path = caller_path

    def find_all_files(self, filename: str) -> t.List[Path]:
        """
        Принимает название или часть названия файла
        и возвращает список подходящих в порядке
        более глубокой вложенности
        """
        paths = []
        depth_limit = 4
        curr_dirname = os.path.dirname(self._caller_path)
        curr_dirname = Path(curr_dirname)
        for i in range(depth_limit):
            path = self._get_path(str(curr_dirname), filename)
            if path:
                paths.append(path)
            if curr_dirname == self._root_path:
                break

            curr_dirname = curr_dirname.parent

        # Самый последний, наиболее глубокий в файловой структуре
        return paths[::-1]

    @classmethod
    def _get_path(cls, dir_path: str, filename: str) -> t.Optional[Path]:
        """Проверяет в данной директории наличие файла
        по паттерну filename, возвращает None при отсутствии
        и Path() если существует"""
        dir_path = Path(dir_path)
        if not dir_path.exists():
            return None
        files = os.listdir(dir_path)
        for file in files:
            filepath = os.path.join(dir_path, file)
            if filename in filepath:
                return Path(filepath)

        return None


class ConfigSourceAdapter:
    """
    Предоставляет ряд универсальных методов для работы с источником конфига.
    Инкапсулирует реальный доступ к источнику
    Парсинг, модификация, доступ к полям и прочее
    """

    def __init__(self, source: Source):
        self._source = source
        self._adapter = self._adapters_dict[source.source_type]

    def get_dict(self) -> dict:
        return self._adapter.get_dict(self._source)

    _adapters_dict = {
        SourceType.ENV: EnvAdapter,
        SourceType.FILE: FileAdapter,
        SourceType.DICT: DictAdapter
    }


class ConfigAggregator:
    """Превращает сырые словари из файлов и других источников
     в итоговый набор конфигов для пользования"""

    def __init__(self, source: t.List[Source]):
        self._sources = source

    def to_dict(self) -> dict:
        """Возвращает готовый итоговый словарь, содержащий
        все необходимые данные (переменные конфигурации)"""
        return self._combine_sources()

    @classmethod
    def _extract_source(cls, source) -> dict:
        """Возвращает соответствующий источнику словарь данных"""
        adapter = ConfigSourceAdapter(source)
        return adapter.get_dict()

    def _combine_sources(self) -> dict:
        """Возвращает общий для всех источников словарь"""
        data = {}
        for source in self._sources:
            data.update(self._extract_source(source))

        return data



from enum import Enum, auto
from pathlib import Path


class SourceType(Enum):
    """Тип переданного в Config() источника """
    ENV = auto()
    DICT = auto()
    FILE = auto()


class Source:
    """Класс обертка над источникоми конфигов,
    например:
     - Source.env - указывает на переменные окружения
     TODO: Дописать примеры использования
     """
    env = '__ENV__'

    def __init__(self, value, manual=False):
        """
        :param value:
        :param manual: True, если данный источник указал пользователь вручную
        False, если взято из настроек по умолчанию
        """
        # TODO точка роста, сюда можно добавить другие источники

        # dict в котором хранятся данные, характеризующие источник
        self.data = {}

        if value == self.env:
            self.source_type = SourceType.ENV
        elif isinstance(value, dict):
            self.source_type = SourceType.DICT
            self.data = value
        elif isinstance(value, str):
            self.source_type = SourceType.FILE
            self.data['filename'] = value
        elif isinstance(value, Path):
            self.source_type = SourceType.FILE
            self.data['filepath'] = value
        else:
            raise ValueError('Unknown source type %s' % type(value))

        self.manual = manual

    def __repr__(self):
        return f'Source({self.data})' if self.data else f'Source(type={self.source_type})'
